@@ -99,6 +99,8 @@ struct resolv_nameserver {
 struct _GResolv {
 	gint ref_count;
 
+	int result_family;
+
 	guint next_lookup_id;
 	GQueue *lookup_queue;
 	GQueue *query_queue;
@@ -562,6 +564,8 @@ static int send_query(GResolv *resolv, const unsigned char *buf, int len)
 		sk = g_io_channel_unix_get_fd(nameserver->udp_channel);
 
 		sent = send(sk, buf, len, 0);
+		if (sent < 0)
+			continue;
 	}
 
 	return 0;
@@ -623,7 +627,7 @@ static void parse_response(struct resolv_nameserver *nameserver,
 	GList *list;
 	ns_msg msg;
 	ns_rr rr;
-	int i, n, rcode, count;
+	int i, rcode, count;
 
 	debug(resolv, "response from %s", nameserver->address);
 
@@ -675,7 +679,7 @@ static void parse_response(struct resolv_nameserver *nameserver,
 		lookup->ipv4_query = NULL;
 	}
 
-	for (i = 0, n = 0; i < count; i++) {
+	for (i = 0; i < count; i++) {
 		ns_parserr(&msg, ns_s_an, i, &rr);
 
 		if (ns_rr_class(rr) != ns_c_in)
@@ -789,6 +793,8 @@ GResolv *g_resolv_new(int index)
 
 	resolv->ref_count = 1;
 
+	resolv->result_family = AF_UNSPEC;
+
 	resolv->next_lookup_id = 1;
 
 	resolv->query_queue = g_queue_new();
@@ -908,8 +914,10 @@ static gint add_query(struct resolv_lookup *lookup, const char *hostname, int ty
 
 	query->msgid = buf[0] << 8 | buf[1];
 
-	if (send_query(lookup->resolv, buf, len) < 0)
+	if (send_query(lookup->resolv, buf, len) < 0) {
+		g_free(query);
 		return -EIO;
+	}
 
 	query->resolv = lookup->resolv;
 	query->lookup = lookup;
@@ -970,16 +978,24 @@ guint g_resolv_lookup_hostname(GResolv *resolv, const char *hostname,
 	lookup->result_data = user_data;
 	lookup->id = resolv->next_lookup_id++;
 
-	if (add_query(lookup, hostname, ns_t_a)) {
-		g_free(lookup);
-		return -EIO;
+	if (resolv->result_family != AF_INET6) {
+		if (add_query(lookup, hostname, ns_t_a)) {
+			g_free(lookup);
+			return -EIO;
+		}
 	}
 
-	if (add_query(lookup, hostname, ns_t_aaaa)) {
-		destroy_query(lookup->ipv4_query);
-		g_queue_remove(resolv->query_queue, lookup->ipv4_query);
-		g_free(lookup);
-		return -EIO;
+	if (resolv->result_family != AF_INET) {
+		if (add_query(lookup, hostname, ns_t_aaaa)) {
+			if (resolv->result_family != AF_INET6) {
+				destroy_query(lookup->ipv4_query);
+				g_queue_remove(resolv->query_queue,
+						lookup->ipv4_query);
+			}
+
+			g_free(lookup);
+			return -EIO;
+		}
 	}
 
 	g_queue_push_tail(resolv->lookup_queue, lookup);
@@ -1001,3 +1017,16 @@ gboolean g_resolv_cancel_lookup(GResolv *resolv, guint id)
 
 	return TRUE;
 }
+
+gboolean g_resolv_set_address_family(GResolv *resolv, int family)
+{
+	if (resolv == NULL)
+		return FALSE;
+
+	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6)
+		return FALSE;
+
+	resolv->result_family = family;
+
+	return TRUE;
+}
