@@ -45,103 +45,6 @@ static in_addr_t loopback_netmask;
 
 static char system_hostname[HOST_NAME_MAX + 1];
 
-#if 0
-static GIOChannel *inotify_channel = NULL;
-
-static int hostname_descriptor = -1;
-
-static gboolean inotify_event(GIOChannel *channel,
-					GIOCondition condition, gpointer data)
-{
-	unsigned char buf[129], *ptr = buf;
-	gsize len;
-	GIOError err;
-
-	if (condition & (G_IO_HUP | G_IO_ERR))
-		return FALSE;
-
-	memset(buf, 0, sizeof(buf));
-
-	err = g_io_channel_read(channel, (gchar *) buf, sizeof(buf) - 1, &len);
-	if (err != G_IO_ERROR_NONE) {
-		if (err == G_IO_ERROR_AGAIN)
-			return TRUE;
-		connman_error("Reading from inotify channel failed");
-		return FALSE;
-	}
-
-	while (len >= sizeof(struct inotify_event)) {
-		struct inotify_event *evt = (struct inotify_event *) ptr;
-
-		if (evt->wd == hostname_descriptor) {
-			if (evt->mask & (IN_CREATE | IN_MOVED_TO))
-				connman_info("create hostname file");
-
-			if (evt->mask & (IN_DELETE | IN_MOVED_FROM))
-				connman_info("delete hostname file");
-
-			if (evt->mask & (IN_MODIFY | IN_MOVE_SELF))
-				connman_info("modify hostname file");
-		}
-
-		len -= sizeof(struct inotify_event) + evt->len;
-		ptr += sizeof(struct inotify_event) + evt->len;
-	}
-
-	return TRUE;
-}
-
-static int create_watch(void)
-{
-	int fd;
-
-	fd = inotify_init();
-	if (fd < 0) {
-		connman_error("Creation of inotify context failed");
-		return -EIO;
-	}
-
-	inotify_channel = g_io_channel_unix_new(fd);
-	if (inotify_channel == NULL) {
-		connman_error("Creation of inotify channel failed");
-		close(fd);
-		return -EIO;
-	}
-
-	hostname_descriptor = inotify_add_watch(fd, "/etc/hostname",
-				IN_MODIFY | IN_DELETE_SELF | IN_MOVE_SELF);
-	if (hostname_descriptor < 0) {
-		connman_error("Creation of hostname watch failed");
-		g_io_channel_unref(inotify_channel);
-		inotify_channel = NULL;
-		close(fd);
-		return -EIO;
-	}
-
-	g_io_add_watch(inotify_channel, G_IO_IN | G_IO_ERR | G_IO_HUP,
-							inotify_event, NULL);
-
-	return 0;
-}
-
-static void remove_watch(void)
-{
-	int fd;
-
-	if (inotify_channel == NULL)
-		return;
-
-	fd = g_io_channel_unix_get_fd(inotify_channel);
-
-	if (hostname_descriptor >= 0)
-		inotify_rm_watch(fd, hostname_descriptor);
-
-	g_io_channel_unref(inotify_channel);
-
-	close(fd);
-}
-#endif
-
 static void create_hostname(void)
 {
 	const char *name = "localhost";
@@ -186,17 +89,26 @@ static gboolean valid_loopback(int sk, struct ifreq *ifr)
 {
 	struct sockaddr_in *addr;
 	int err;
+	char buf[INET_ADDRSTRLEN];
+
+	/* It is possible to end up in situations in which the
+	 * loopback interface is up but has no valid address. In that
+	 * case, we expect EADDRNOTAVAIL and should return FALSE.
+	 */
 
 	err = ioctl(sk, SIOCGIFADDR, ifr);
 	if (err < 0) {
 		err = -errno;
 		connman_error("Getting address failed (%s)", strerror(-err));
-		return TRUE;
+		return err != -EADDRNOTAVAIL ? TRUE : FALSE;
 	}
 
 	addr = (struct sockaddr_in *) &ifr->ifr_addr;
-	if (addr->sin_addr.s_addr != loopback_address)
+	if (addr->sin_addr.s_addr != loopback_address) {
+		connman_warn("Invalid loopback address %s",
+			inet_ntop(AF_INET, &addr->sin_addr, buf, sizeof(buf)));
 		return FALSE;
+	}
 
 	err = ioctl(sk, SIOCGIFNETMASK, ifr);
 	if (err < 0) {
@@ -206,8 +118,11 @@ static gboolean valid_loopback(int sk, struct ifreq *ifr)
 	}
 
 	addr = (struct sockaddr_in *) &ifr->ifr_netmask;
-	if (addr->sin_addr.s_addr != loopback_netmask)
+	if (addr->sin_addr.s_addr != loopback_netmask) {
+		connman_warn("Invalid loopback netmask %s",
+			inet_ntop(AF_INET, &addr->sin_addr, buf, sizeof(buf)));
 		return FALSE;
+	}
 
 	return TRUE;
 }
@@ -220,7 +135,7 @@ static int setup_loopback(void)
 
 	sk = socket(PF_INET, SOCK_DGRAM, 0);
 	if (sk < 0)
-		return -1;
+		return -errno;
 
 	memset(&ifr, 0, sizeof(ifr));
 	strcpy(ifr.ifr_name, "lo");
@@ -242,7 +157,7 @@ static int setup_loopback(void)
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
-	addr.sin_addr.s_addr = inet_addr("127.0.0.1");
+	addr.sin_addr.s_addr = loopback_address;
 	memcpy(&ifr.ifr_addr, &addr, sizeof(ifr.ifr_addr));
 
 	err = ioctl(sk, SIOCSIFADDR, &ifr);
@@ -254,7 +169,7 @@ static int setup_loopback(void)
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
-	addr.sin_addr.s_addr = inet_addr("255.0.0.0");
+	addr.sin_addr.s_addr = loopback_netmask;
 	memcpy(&ifr.ifr_netmask, &addr, sizeof(ifr.ifr_netmask));
 
 	err = ioctl(sk, SIOCSIFNETMASK, &ifr);
@@ -338,8 +253,6 @@ static int loopback_init(void)
 
 	setup_hostname();
 
-	//create_watch();
-
 	connman_utsname_driver_register(&loopback_driver);
 
 	return 0;
@@ -348,8 +261,6 @@ static int loopback_init(void)
 static void loopback_exit(void)
 {
 	connman_utsname_driver_unregister(&loopback_driver);
-
-	//remove_watch();
 }
 
 CONNMAN_PLUGIN_DEFINE(loopback, "Loopback device plugin", VERSION,
