@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <stdio.h>
 #include <net/if.h>
 #include <net/if_arp.h>
@@ -53,6 +54,7 @@ struct connman_ipconfig {
 	struct connman_ipaddress *system;
 
 	int ipv6_privacy_config;
+	char *last_dhcp_address;
 };
 
 struct connman_ipdevice {
@@ -145,6 +147,9 @@ static gboolean check_ipv6_address(const char *address)
 	unsigned char buf[sizeof(struct in6_addr)];
 	int err;
 
+	if (address == NULL)
+		return FALSE;
+
 	err = inet_pton(AF_INET6, address, buf);
 	if (err > 0)
 		return TRUE;
@@ -153,8 +158,9 @@ static gboolean check_ipv6_address(const char *address)
 }
 
 int connman_ipaddress_set_ipv6(struct connman_ipaddress *ipaddress,
-				const char *address, const char *gateway,
-						unsigned char prefix_length)
+				const char *address,
+				unsigned char prefix_length,
+				const char *gateway)
 {
 	if (ipaddress == NULL)
 		return -EINVAL;
@@ -168,6 +174,8 @@ int connman_ipaddress_set_ipv6(struct connman_ipaddress *ipaddress,
 	DBG("prefix_len %d address %s gateway %s",
 			prefix_length, address, gateway);
 
+	ipaddress->family = AF_INET6;
+
 	ipaddress->prefixlen = prefix_length;
 
 	g_free(ipaddress->local);
@@ -179,11 +187,13 @@ int connman_ipaddress_set_ipv6(struct connman_ipaddress *ipaddress,
 	return 0;
 }
 
-void connman_ipaddress_set_ipv4(struct connman_ipaddress *ipaddress,
+int connman_ipaddress_set_ipv4(struct connman_ipaddress *ipaddress,
 		const char *address, const char *netmask, const char *gateway)
 {
 	if (ipaddress == NULL)
-		return;
+		return -EINVAL;
+
+	ipaddress->family = AF_INET;
 
 	ipaddress->prefixlen = __connman_ipconfig_netmask_prefix_len(netmask);
 
@@ -192,6 +202,18 @@ void connman_ipaddress_set_ipv4(struct connman_ipaddress *ipaddress,
 
 	g_free(ipaddress->gateway);
 	ipaddress->gateway = g_strdup(gateway);
+
+	return 0;
+}
+
+void connman_ipaddress_set_peer(struct connman_ipaddress *ipaddress,
+				const char *peer)
+{
+	if (ipaddress == NULL)
+		return;
+
+	g_free(ipaddress->peer);
+	ipaddress->peer = g_strdup(peer);
 }
 
 void connman_ipaddress_clear(struct connman_ipaddress *ipaddress)
@@ -267,6 +289,20 @@ static struct connman_ipaddress *find_ipaddress(struct connman_ipdevice *ipdevic
 	return NULL;
 }
 
+const char *__connman_ipconfig_type2string(enum connman_ipconfig_type type)
+{
+	switch (type) {
+	case CONNMAN_IPCONFIG_TYPE_UNKNOWN:
+		return "unknown";
+	case CONNMAN_IPCONFIG_TYPE_IPV4:
+		return "IPv4";
+	case CONNMAN_IPCONFIG_TYPE_IPV6:
+		return "IPv6";
+	}
+
+	return NULL;
+}
+
 static const char *type2str(unsigned short type)
 {
 	switch (type) {
@@ -673,6 +709,7 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 {
 	struct connman_ipdevice *ipdevice;
 	struct connman_ipaddress *ipaddress;
+	enum connman_ipconfig_type type;
 	GList *list;
 
 	DBG("index %d", index);
@@ -694,6 +731,13 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 		return;
 	}
 
+	if (family == AF_INET)
+		type = CONNMAN_IPCONFIG_TYPE_IPV4;
+	else if (family == AF_INET6)
+		type = CONNMAN_IPCONFIG_TYPE_IPV6;
+	else
+		return;
+
 	ipdevice->address_list = g_slist_append(ipdevice->address_list,
 								ipaddress);
 
@@ -720,6 +764,9 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 		if (index != ipconfig->index)
 			continue;
 
+		if (type != ipconfig->type)
+			continue;
+
 		if (ipconfig->ops == NULL)
 			continue;
 
@@ -733,6 +780,7 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 {
 	struct connman_ipdevice *ipdevice;
 	struct connman_ipaddress *ipaddress;
+	enum connman_ipconfig_type type;
 	GList *list;
 
 	DBG("index %d", index);
@@ -745,6 +793,13 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 	if (ipaddress == NULL)
 		return;
 
+	if (family == AF_INET)
+		type = CONNMAN_IPCONFIG_TYPE_IPV4;
+	else if (family == AF_INET6)
+		type = CONNMAN_IPCONFIG_TYPE_IPV6;
+	else
+		return;
+
 	ipdevice->address_list = g_slist_remove(ipdevice->address_list,
 								ipaddress);
 
@@ -767,6 +822,9 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 		if (index != ipconfig->index)
 			continue;
 
+		if (type != ipconfig->type)
+			continue;
+
 		if (ipconfig->ops == NULL)
 			continue;
 
@@ -786,11 +844,14 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 	if (ipdevice == NULL)
 		return;
 
-	if (scope == 0 && g_strcmp0(dst, "0.0.0.0") == 0) {
+	if (scope == 0 && (g_strcmp0(dst, "0.0.0.0") == 0 ||
+						g_strcmp0(dst, "::") == 0)) {
 		GSList *list;
 		GList *config_list;
+		enum connman_ipconfig_type type;
 
 		if (family == AF_INET6) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV6;
 			g_free(ipdevice->ipv6_gateway);
 			ipdevice->ipv6_gateway = g_strdup(gateway);
 
@@ -800,7 +861,8 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 				ipdevice->config_ipv6->system->gateway =
 					g_strdup(gateway);
 			}
-		} else {
+		} else if (family == AF_INET) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV4;
 			g_free(ipdevice->ipv4_gateway);
 			ipdevice->ipv4_gateway = g_strdup(gateway);
 
@@ -810,7 +872,8 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 				ipdevice->config_ipv4->system->gateway =
 					g_strdup(gateway);
 			}
-		}
+		} else
+			return;
 
 		for (list = ipdevice->address_list; list; list = list->next) {
 			struct connman_ipaddress *ipaddress = list->data;
@@ -826,6 +889,9 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 			if (index != ipconfig->index)
 				continue;
 
+			if (type != ipconfig->type)
+				continue;
+
 			if (ipconfig->ops == NULL)
 				continue;
 
@@ -850,11 +916,14 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 	if (ipdevice == NULL)
 		return;
 
-	if (scope == 0 && g_strcmp0(dst, "0.0.0.0") == 0) {
+	if (scope == 0 && (g_strcmp0(dst, "0.0.0.0") == 0 ||
+						g_strcmp0(dst, "::") == 0)) {
 		GSList *list;
 		GList *config_list;
+		enum connman_ipconfig_type type;
 
 		if (family == AF_INET6) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV6;
 			g_free(ipdevice->ipv6_gateway);
 			ipdevice->ipv6_gateway = NULL;
 
@@ -863,7 +932,8 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 				g_free(ipdevice->config_ipv6->system->gateway);
 				ipdevice->config_ipv6->system->gateway = NULL;
 			}
-		} else {
+		} else if (family == AF_INET) {
+			type = CONNMAN_IPCONFIG_TYPE_IPV4;
 			g_free(ipdevice->ipv4_gateway);
 			ipdevice->ipv4_gateway = NULL;
 
@@ -872,7 +942,8 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 				g_free(ipdevice->config_ipv4->system->gateway);
 				ipdevice->config_ipv4->system->gateway = NULL;
 			}
-		}
+		} else
+			return;
 
 		for (list = ipdevice->address_list; list; list = list->next) {
 			struct connman_ipaddress *ipaddress = list->data;
@@ -888,6 +959,9 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 			if (index != ipconfig->index)
 				continue;
 
+			if (type != ipconfig->type)
+				continue;
+
 			if (ipconfig->ops == NULL)
 				continue;
 
@@ -1038,31 +1112,51 @@ const char *__connman_ipconfig_get_gateway(struct connman_ipconfig *ipconfig)
 
 void __connman_ipconfig_set_gateway(struct connman_ipconfig *ipconfig, const char *gateway)
 {
-	struct connman_service *service;
+	DBG("");
 
 	if (ipconfig->address == NULL)
 		return;
+	g_free(ipconfig->address->gateway);
+	ipconfig->address->gateway = g_strdup(gateway);
+}
+
+int __connman_ipconfig_gateway_add(struct connman_ipconfig *ipconfig)
+{
+	struct connman_service *service;
+
+	DBG("");
+
+	if (ipconfig->address == NULL)
+		return -EINVAL;
+
+	service = __connman_service_lookup_from_index(ipconfig->index);
+	if (service == NULL)
+		return -EINVAL;
+
+	__connman_connection_gateway_remove(service, ipconfig->type);
+
+	DBG("type %d gw %s peer %s", ipconfig->type,
+		ipconfig->address->gateway, ipconfig->address->peer);
+
+	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6 ||
+				ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		return __connman_connection_gateway_add(service,
+						ipconfig->address->gateway,
+						ipconfig->type,
+						ipconfig->address->peer);
+
+	return 0;
+}
+
+void __connman_ipconfig_gateway_remove(struct connman_ipconfig *ipconfig)
+{
+	struct connman_service *service;
+
+	DBG("");
 
 	service = __connman_service_lookup_from_index(ipconfig->index);
 	if (service != NULL)
-		__connman_connection_gateway_remove(service);
-
-	g_free(ipconfig->address->gateway);
-	ipconfig->address->gateway = g_strdup(gateway);
-
-	if (service != NULL && ipconfig->address->gateway != NULL) {
-		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6) {
-			__connman_connection_gateway_add(service,
-						NULL,
-						ipconfig->address->gateway,
-						ipconfig->address->peer);
-		} else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4) {
-			__connman_connection_gateway_add(service,
-						ipconfig->address->gateway,
-						NULL,
-						ipconfig->address->peer);
-		}
-	}
+		__connman_connection_gateway_remove(service, ipconfig->type);
 }
 
 unsigned char __connman_ipconfig_get_prefixlen(struct connman_ipconfig *ipconfig)
@@ -1193,6 +1287,7 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
 
 		connman_ipaddress_free(ipconfig->system);
 		connman_ipaddress_free(ipconfig->address);
+		g_free(ipconfig->last_dhcp_address);
 		g_free(ipconfig);
 	}
 }
@@ -1300,44 +1395,17 @@ enum connman_ipconfig_method __connman_ipconfig_get_method(struct connman_ipconf
 	return ipconfig->method;
 }
 
-/**
- * connman_ipconfig_bind:
- * @ipconfig: ipconfig structure
- * @ipaddress: ipaddress structure
- *
- * Bind IP address details to configuration
- */
-void connman_ipconfig_bind(struct connman_ipconfig *ipconfig,
-					struct connman_ipaddress *ipaddress)
-{
-	struct connman_ipconfig *origin;
-
-	origin = ipconfig->origin ? ipconfig->origin : ipconfig;
-
-	connman_ipaddress_copy(origin->address, ipaddress);
-
-	connman_inet_set_address(origin->index, origin->address);
-}
-
-void __connman_ipconfig_set_element_ipv6_gateway(
-			struct connman_ipconfig *ipconfig,
-				struct connman_element *element)
-{
-	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
-		element->ipv6.gateway = ipconfig->address->gateway;
-}
-
-int __connman_ipconfig_set_address(struct connman_ipconfig *ipconfig)
+int __connman_ipconfig_address_add(struct connman_ipconfig *ipconfig)
 {
 	DBG("");
 
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_FIXED:
-	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
 			return connman_inet_set_address(ipconfig->index,
@@ -1350,8 +1418,10 @@ int __connman_ipconfig_set_address(struct connman_ipconfig *ipconfig)
 	return 0;
 }
 
-int __connman_ipconfig_clear_address(struct connman_ipconfig *ipconfig)
+int __connman_ipconfig_address_remove(struct connman_ipconfig *ipconfig)
 {
+	int err;
+
 	DBG("");
 
 	if (ipconfig == NULL)
@@ -1362,19 +1432,51 @@ int __connman_ipconfig_clear_address(struct connman_ipconfig *ipconfig)
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_FIXED:
-	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		err = __connman_ipconfig_address_unset(ipconfig);
+		connman_ipaddress_clear(ipconfig->address);
+
+		return err;
+	}
+
+	return 0;
+}
+
+int __connman_ipconfig_address_unset(struct connman_ipconfig *ipconfig)
+{
+	int err;
+
+	DBG("");
+
+	if (ipconfig == NULL)
+		return 0;
+
+	DBG("method %d", ipconfig->method);
+
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
+		break;
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
-			return connman_inet_clear_address(ipconfig->index,
+			err = connman_inet_clear_address(ipconfig->index,
 							ipconfig->address);
 		else if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6)
-			return connman_inet_clear_ipv6_address(
+			err = connman_inet_clear_ipv6_address(
 						ipconfig->index,
 						ipconfig->address->local,
 						ipconfig->address->prefixlen);
+		else
+			err = -EINVAL;
+
+		return err;
 	}
 
 	return 0;
@@ -1418,6 +1520,24 @@ const char *__connman_ipconfig_get_proxy_autoconfig(struct connman_ipconfig *ipc
 	return ipdevice->pac;
 }
 
+void __connman_ipconfig_set_dhcp_address(struct connman_ipconfig *ipconfig,
+					const char *address)
+{
+	if (ipconfig == NULL)
+		return;
+
+	g_free(ipconfig->last_dhcp_address);
+	ipconfig->last_dhcp_address = g_strdup(address);
+}
+
+char *__connman_ipconfig_get_dhcp_address(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL)
+		return NULL;
+
+	return ipconfig->last_dhcp_address;
+}
+
 static void disable_ipv6(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
@@ -1450,6 +1570,14 @@ static void enable_ipv6(struct connman_ipconfig *ipconfig)
 	set_ipv6_state(ipdevice->ifname, TRUE);
 }
 
+void __connman_ipconfig_enable_ipv6(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL || ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV6)
+		return;
+
+	enable_ipv6(ipconfig);
+}
+
 void __connman_ipconfig_disable_ipv6(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL || ipconfig->type != CONNMAN_IPCONFIG_TYPE_IPV6)
@@ -1680,7 +1808,8 @@ void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
 }
 
 void __connman_ipconfig_append_ipv6(struct connman_ipconfig *ipconfig,
-							DBusMessageIter *iter)
+					DBusMessageIter *iter,
+					struct connman_ipconfig *ipconfig_ipv4)
 {
 	const char *str, *privacy;
 
@@ -1693,6 +1822,12 @@ void __connman_ipconfig_append_ipv6(struct connman_ipconfig *ipconfig,
 	if (str == NULL)
 		return;
 
+	if (ipconfig_ipv4 != NULL &&
+			ipconfig->method == CONNMAN_IPCONFIG_METHOD_AUTO) {
+		if (__connman_6to4_check(ipconfig_ipv4) == 1)
+			str = "6to4";
+	}
+
 	connman_dbus_dict_append_basic(iter, "Method", DBUS_TYPE_STRING, &str);
 
 	if (ipconfig->system == NULL)
@@ -1919,7 +2054,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 		else
 			return connman_ipaddress_set_ipv6(
 					ipconfig->address, address,
-						gateway, prefix_length);
+						prefix_length, gateway);
 		break;
 
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
@@ -1965,6 +2100,7 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 {
 	char *method;
 	char *key;
+	char *str;
 
 	DBG("ipconfig %p identifier %s", ipconfig, identifier);
 
@@ -2025,6 +2161,14 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 				keyfile, identifier, key, NULL);
 	g_free(key);
 
+	key = g_strdup_printf("%sDHCP.LastAddress", prefix);
+	str = g_key_file_get_string(keyfile, identifier, key, NULL);
+	if (str != NULL) {
+		g_free(ipconfig->last_dhcp_address);
+		ipconfig->last_dhcp_address = str;
+	}
+	g_free(key);
+
 	return 0;
 }
 
@@ -2042,6 +2186,34 @@ int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 	g_key_file_set_string(keyfile, identifier, key, method);
 	g_free(key);
 
+	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6) {
+		const char *privacy;
+		privacy = privacy2string(ipconfig->ipv6_privacy_config);
+		key = g_strdup_printf("%sprivacy", prefix);
+		g_key_file_set_string(keyfile, identifier, key, privacy);
+		g_free(key);
+	}
+
+	switch (ipconfig->method) {
+	case CONNMAN_IPCONFIG_METHOD_FIXED:
+	case CONNMAN_IPCONFIG_METHOD_MANUAL:
+		break;
+	case CONNMAN_IPCONFIG_METHOD_DHCP:
+		key = g_strdup_printf("%sDHCP.LastAddress", prefix);
+		if (ipconfig->last_dhcp_address != NULL &&
+				strlen(ipconfig->last_dhcp_address) > 0)
+			g_key_file_set_string(keyfile, identifier, key,
+					ipconfig->last_dhcp_address);
+		else
+			g_key_file_remove_key(keyfile, identifier, key, NULL);
+		g_free(key);
+		/* fall through */
+	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
+	case CONNMAN_IPCONFIG_METHOD_OFF:
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
+		return 0;
+	}
+
 	key = g_strdup_printf("%snetmask_prefixlen", prefix);
 	g_key_file_set_integer(keyfile, identifier,
 			key, ipconfig->address->prefixlen);
@@ -2071,14 +2243,6 @@ int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
 			key, ipconfig->address->gateway);
 	g_free(key);
 
-	if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV6) {
-		const char *privacy;
-		privacy = privacy2string(ipconfig->ipv6_privacy_config);
-		key = g_strdup_printf("%sprivacy", prefix);
-		g_key_file_set_string(keyfile, identifier, key, privacy);
-		g_free(key);
-	}
-
 	return 0;
 }
 
