@@ -30,6 +30,7 @@
 static DBusConnection *connection = NULL;
 
 static GSList *notifier_list = NULL;
+static GHashTable *service_hash = NULL;
 
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
@@ -241,6 +242,11 @@ void __connman_notifier_unregister(enum connman_service_type type)
 {
 	DBG("type %d", type);
 
+	if (g_atomic_int_get(&registered[type]) == 0) {
+		connman_error("notifier unregister underflow");
+		return;
+	}
+
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
@@ -287,6 +293,11 @@ void __connman_notifier_disable(enum connman_service_type type)
 {
 	DBG("type %d", type);
 
+	if (g_atomic_int_get(&enabled[type]) == 0) {
+		connman_error("notifier disable underflow");
+		return;
+	}
+
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
@@ -333,6 +344,11 @@ void __connman_notifier_disconnect(enum connman_service_type type)
 {
 	DBG("type %d", type);
 
+	if (g_atomic_int_get(&connected[type]) == 0) {
+		connman_error("notifier disconnect underflow");
+		return;
+	}
+
 	switch (type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
@@ -385,7 +401,8 @@ void __connman_notifier_default_changed(struct connman_service *service)
 	}
 }
 
-void __connman_notifier_service_add(struct connman_service *service)
+void __connman_notifier_service_add(struct connman_service *service,
+					const char *name)
 {
 	GSList *list;
 
@@ -393,7 +410,7 @@ void __connman_notifier_service_add(struct connman_service *service)
 		struct connman_notifier *notifier = list->data;
 
 		if (notifier->service_add)
-			notifier->service_add(service);
+			notifier->service_add(service, name);
 	}
 }
 
@@ -401,6 +418,17 @@ void __connman_notifier_service_remove(struct connman_service *service)
 {
 	GSList *list;
 
+	if (g_hash_table_lookup(service_hash, service) != NULL) {
+		/*
+		 * This is a tempory check for consistency. It can be
+		 * removed when there are no reports for the following
+		 * error message.
+		 */
+		connman_error("Service state machine inconsistency detected.");
+
+		g_hash_table_remove(service_hash, service);
+	}
+
 	for (list = notifier_list; list; list = list->next) {
 		struct connman_notifier *notifier = list->data;
 
@@ -448,10 +476,26 @@ void __connman_notifier_offlinemode(connman_bool_t enabled)
 	}
 }
 
+static void notify_idle_state(connman_bool_t idle)
+{
+	GSList *list;
+
+	DBG("idle %d", idle);
+
+	for (list = notifier_list; list; list = list->next) {
+		struct connman_notifier *notifier = list->data;
+
+		if (notifier->idle_state)
+			notifier->idle_state(idle);
+	}
+}
+
 void __connman_notifier_service_state_changed(struct connman_service *service,
 					enum connman_service_state state)
 {
 	GSList *list;
+	unsigned int old_size;
+	connman_bool_t found;
 
 	for (list = notifier_list; list; list = list->next) {
 		struct connman_notifier *notifier = list->data;
@@ -459,6 +503,36 @@ void __connman_notifier_service_state_changed(struct connman_service *service,
 		if (notifier->service_state_changed)
 			notifier->service_state_changed(service, state);
 	}
+
+	old_size = g_hash_table_size(service_hash);
+	found = g_hash_table_lookup(service_hash, service) != NULL;
+
+	switch (state) {
+	case CONNMAN_SERVICE_STATE_UNKNOWN:
+	case CONNMAN_SERVICE_STATE_DISCONNECT:
+	case CONNMAN_SERVICE_STATE_IDLE:
+		if (found == FALSE)
+			break;
+
+		g_hash_table_remove(service_hash, service);
+		if (old_size == 1)
+			notify_idle_state(TRUE);
+
+		break;
+	case CONNMAN_SERVICE_STATE_ASSOCIATION:
+	case CONNMAN_SERVICE_STATE_CONFIGURATION:
+	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_ONLINE:
+	case CONNMAN_SERVICE_STATE_FAILURE:
+		if (found == TRUE)
+			break;
+
+		g_hash_table_insert(service_hash, service, service);
+		if (old_size == 0)
+			notify_idle_state(FALSE);
+
+		break;
+	}
 }
 
 void __connman_notifier_ipconfig_changed(struct connman_service *service,
@@ -526,6 +600,10 @@ int __connman_notifier_init(void)
 
 	connection = connman_dbus_get_connection();
 
+	service_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+						NULL, NULL);
+
+
 	return 0;
 }
 
@@ -533,5 +611,8 @@ void __connman_notifier_cleanup(void)
 {
 	DBG("");
 
+	g_hash_table_destroy(service_hash);
+	service_hash = NULL;
+
 	dbus_connection_unref(connection);
 }
