@@ -76,13 +76,14 @@ static const char *hooknames[] = {
 
 #define ALIGN(s) (((s) + ((MIN_ALIGN)-1)) & ~((MIN_ALIGN)-1))
 
-struct ipt_error_target {
+struct error_target {
 	struct xt_entry_target t;
 	char error[IPT_TABLE_MAXNAMELEN];
 };
 
 struct connman_iptables_entry {
 	int offset;
+	int builtin;
 
 	struct ipt_entry *entry;
 };
@@ -97,6 +98,9 @@ struct connman_iptables {
 	unsigned int old_entries;
 	unsigned int size;
 
+	unsigned int underflow[NF_INET_NUMHOOKS];
+	unsigned int hook_entry[NF_INET_NUMHOOKS];
+
 	GList *entries;
 };
 
@@ -186,13 +190,11 @@ static gboolean is_jump(struct connman_iptables_entry *e)
 static gboolean is_chain(struct connman_iptables *table,
 				struct connman_iptables_entry *e)
 {
-	int builtin;
 	struct ipt_entry *entry;
 	struct xt_entry_target *target;
 
 	entry = e->entry;
-	builtin = is_hook_entry(table, entry);
-	if (builtin >= 0)
+	if (e->builtin >= 0)
 		return TRUE;
 
 	target = ipt_get_target(entry);
@@ -216,7 +218,7 @@ static GList *find_chain_head(struct connman_iptables *table,
 		entry = head->entry;
 
 		/* Buit-in chain */
-		builtin = is_hook_entry(table, entry);
+		builtin = head->builtin;
 		if (builtin >= 0 && !strcmp(hooknames[builtin], chain_name))
 			break;
 
@@ -245,7 +247,7 @@ static GList *find_chain_tail(struct connman_iptables *table,
 		entry = head->entry;
 
 		/* Buit-in chain */
-		builtin = is_hook_entry(table, entry);
+		builtin = head->builtin;
 		if (builtin >= 0 && !strcmp(hooknames[builtin], chain_name))
 			break;
 
@@ -297,7 +299,8 @@ static void update_offsets(struct connman_iptables *table)
 }
 
 static int connman_add_entry(struct connman_iptables *table,
-				struct ipt_entry *entry, GList *before)
+				struct ipt_entry *entry, GList *before,
+					int builtin)
 {
 	GList *list;
 	struct connman_iptables_entry *e, *tmp, *entry_before;
@@ -311,6 +314,7 @@ static int connman_add_entry(struct connman_iptables *table,
 		return -1;
 
 	e->entry = entry;
+	e->builtin = builtin;
 
 	table->entries = g_list_insert_before(table->entries, before, e);
 	table->num_entries++;
@@ -345,6 +349,71 @@ static int connman_add_entry(struct connman_iptables *table,
 	return 0;
 }
 
+static int connman_iptables_flush_chain(struct connman_iptables *table,
+						char *name)
+{
+	GList *chain_head, *chain_tail, *list, *next;
+	struct connman_iptables_entry *entry;
+	int builtin, removed = 0;
+
+	chain_head = find_chain_head(table, name);
+	if (chain_head == NULL)
+		return -EINVAL;
+
+	chain_tail = find_chain_tail(table, name);
+	if (chain_tail == NULL)
+		return -EINVAL;
+
+	entry = chain_head->data;
+	builtin = entry->builtin;
+
+	if (builtin >= 0)
+		list = chain_head;
+	else
+		list = chain_head->next;
+
+	if (list == chain_tail->prev)
+		return 0;
+
+	while (list != chain_tail->prev) {
+		entry = list->data;
+		next = g_list_next(list);
+
+		table->num_entries--;
+		table->size -= entry->entry->next_offset;
+		removed += entry->entry->next_offset;
+
+		table->entries = g_list_remove(table->entries, list->data);
+
+		list = next;
+	}
+
+	if (builtin >= 0) {
+		struct connman_iptables_entry *e;
+
+		entry = list->data;
+
+		entry->builtin = builtin;
+
+		table->underflow[builtin] -= removed;
+
+		for (list = chain_tail; list; list = list->next) {
+			e = list->data;
+
+			builtin = e->builtin;
+			if (builtin < 0)
+				continue;
+
+			table->hook_entry[builtin] -= removed;
+			table->underflow[builtin] -= removed;
+		}
+	}
+
+	update_offsets(table);
+
+	return 0;
+}
+
 static int connman_iptables_delete_chain(struct connman_iptables *table,
 						char *name)
 {
@@ -356,7 +425,7 @@ static int connman_iptables_delete_chain(struct connman_iptables *table,
 		return -EINVAL;
 
 	chain_tail = find_chain_tail(table, name);
-	if (chain_head == NULL)
+	if (chain_tail == NULL)
 		return -EINVAL;
 
 	list = chain_head;
@@ -384,7 +453,7 @@ static int connman_iptables_add_chain(struct connman_iptables *table,
 	GList *last;
 	struct ipt_entry *entry_head;
 	struct ipt_entry *entry_return;
-	struct ipt_error_target *error;
+	struct error_target *error;
 	struct ipt_standard_target *standard;
 	u_int16_t entry_head_size, entry_return_size;
 
@@ -401,23 +470,23 @@ static int connman_iptables_add_chain(struct connman_iptables *table,
 
 	/* head entry */
 	entry_head_size = sizeof(struct ipt_entry) +
-				sizeof(struct ipt_error_target);
+				sizeof(struct error_target);
 	entry_head = g_try_malloc0(entry_head_size);
 	if (entry_head == NULL)
-		goto err;
+		goto err_head;
 
 	memset(entry_head, 0, entry_head_size);
 
 	entry_head->target_offset = sizeof(struct ipt_entry);
 	entry_head->next_offset = entry_head_size;
 
-	error = (struct ipt_error_target *) entry_head->elems;
+	error = (struct error_target *) entry_head->elems;
 	strcpy(error->t.u.user.name, IPT_ERROR_TARGET);
-	error->t.u.user.target_size = ALIGN(sizeof(struct ipt_error_target));
+	error->t.u.user.target_size = ALIGN(sizeof(struct error_target));
 	strcpy(error->error, name);
 
-	if (connman_add_entry(table, entry_head, last) < 0)
-		goto err;
+	if (connman_add_entry(table, entry_head, last, -1) < 0)
+		goto err_head;
 
 	/* tail entry */
 	entry_return_size = sizeof(struct ipt_entry) +
@@ -436,14 +505,15 @@ static int connman_iptables_add_chain(struct connman_iptables *table,
 				ALIGN(sizeof(struct ipt_standard_target));
 	standard->verdict = XT_RETURN;
 
-	if (connman_add_entry(table, entry_return, last) < 0)
+	if (connman_add_entry(table, entry_return, last, -1) < 0)
 		goto err;
 
 	return 0;
 
 err:
-	g_free(entry_head);
 	g_free(entry_return);
+err_head:
+	g_free(entry_head);
 
 	return -ENOMEM;
 }
@@ -524,13 +594,48 @@ new_rule(struct connman_iptables *table,
 	return new_entry;
 }
 
+static void update_hooks(struct connman_iptables *table, GList *chain_head, struct ipt_entry *entry)
+{
+	GList *list;
+	struct connman_iptables_entry *head, *e;
+	int builtin;
+
+	if (chain_head == NULL)
+		return;
+
+	head = chain_head->data;
+
+	builtin = head->builtin;
+	if (builtin < 0)
+		return;
+
+	table->underflow[builtin] += entry->next_offset;
+
+	for (list = chain_head->next; list; list = list->next) {
+		e = list->data;
+
+		builtin = e->builtin;
+		if (builtin < 0)
+			continue;
+
+		table->hook_entry[builtin] += entry->next_offset;
+		table->underflow[builtin] += entry->next_offset;
+	}
+}
+
 static int
 connman_iptables_add_rule(struct connman_iptables *table, char *chain_name,
 				char *target_name, struct xtables_target *xt_t,
 				char *match_name, struct xtables_match *xt_m)
 {
-	GList *chain_tail;
+	GList *chain_tail, *chain_head;
 	struct ipt_entry *new_entry;
+	struct connman_iptables_entry *head;
+	int builtin = -1;
+
+	chain_head = find_chain_head(table, chain_name);
+	if (chain_head == NULL)
+		return -EINVAL;
 
 	chain_tail = find_chain_tail(table, chain_name);
 	if (chain_tail == NULL)
@@ -542,7 +647,22 @@ connman_iptables_add_rule(struct connman_iptables *table, char *chain_name,
 	if (new_entry == NULL)
 		return -EINVAL;
 
-	return connman_add_entry(table, new_entry, chain_tail->prev);
+	update_hooks(table, chain_head, new_entry);
+
+	/*
+	 * If the chain is builtin, and does not have any rule,
+	 * then the one that we're inserting is becoming the head
+	 * and thus needs the builtin flag.
+	 */
+	head = chain_head->data;
+	if (head->builtin < 0)
+		builtin = -1;
+	else if (chain_head == chain_tail->prev) {
+		builtin = head->builtin;
+		head->builtin = -1;
+	}
+
+	return connman_add_entry(table, new_entry, chain_tail->prev, builtin);
 }
 
 static struct ipt_replace *
@@ -560,7 +680,7 @@ connman_iptables_blob(struct connman_iptables *table)
 	memset(r, 0, sizeof(*r) + table->size);
 
 	r->counters = g_try_malloc0(sizeof(struct xt_counters)
-				* table->num_entries);
+				* table->old_entries);
 	if (r->counters == NULL) {
 		g_free(r);
 		return NULL;
@@ -573,10 +693,8 @@ connman_iptables_blob(struct connman_iptables *table)
 	r->num_counters = table->old_entries;
 	r->valid_hooks  = table->info->valid_hooks;
 
-	memcpy(r->hook_entry, table->info->hook_entry,
-				sizeof(table->info->hook_entry));
-	memcpy(r->underflow, table->info->underflow,
-				sizeof(table->info->underflow));
+	memcpy(r->hook_entry, table->hook_entry, sizeof(table->hook_entry));
+	memcpy(r->underflow, table->underflow, sizeof(table->underflow));
 
 	entry_index = (unsigned char *)r->entries;
 	for (list = table->entries; list; list = list->next) {
@@ -723,6 +841,22 @@ static int connman_iptables_dump_entry(struct ipt_entry *entry,
 	return 0;
 }
 
+static void connman_iptables_dump_hook(struct connman_iptables *table)
+{
+	int i;
+	printf("hooks: \n");
+	for (i = 0; i < NF_INET_NUMHOOKS; i++) {
+		if ((table->info->valid_hooks & (1 << i)))
+			printf("%s entry %p underflow %p (%#x)\n",
+				hooknames[i],
+				table->blob_entries->entrytable +
+						table->info->hook_entry[i],
+				table->blob_entries->entrytable +
+						table->info->underflow[i],
+					table->info->underflow[i]);
+	}
+}
+
 static void connman_iptables_dump(struct connman_iptables *table)
 {
 	printf("%s valid_hooks=0x%08x, num_entries=%u, size=%u\n",
@@ -730,6 +864,8 @@ static void connman_iptables_dump(struct connman_iptables *table)
 		table->info->valid_hooks, table->info->num_entries,
 		table->info->size);
 
+	connman_iptables_dump_hook(table);
+
 	ENTRY_ITERATE(table->blob_entries->entrytable,
 			table->blob_entries->size,
 			connman_iptables_dump_entry, table);
@@ -785,6 +921,7 @@ static int connman_iptables_commit(struct connman_iptables *table)
 static int add_entry(struct ipt_entry *entry, struct connman_iptables *table)
 {
 	struct ipt_entry *new_entry;
+	int builtin;
 
 	new_entry = g_try_malloc0(entry->next_offset);
 	if (new_entry == NULL)
@@ -792,7 +929,9 @@ static int add_entry(struct ipt_entry *entry, struct connman_iptables *table)
 
 	memcpy(new_entry, entry, entry->next_offset);
 
-	return connman_add_entry(table, new_entry, NULL);
+	builtin = is_hook_entry(table, entry);
+
+	return connman_add_entry(table, new_entry, NULL, builtin);
 }
 
 static struct connman_iptables *connman_iptables_init(const char *table_name)
@@ -833,6 +972,11 @@ static struct connman_iptables *connman_iptables_init(const char *table_name)
 	table->old_entries = table->info->num_entries;
 	table->size = 0;
 
+	memcpy(table->underflow, table->info->underflow,
+				sizeof(table->info->underflow));
+	memcpy(table->hook_entry, table->info->hook_entry,
+				sizeof(table->info->hook_entry));
+
 	ENTRY_ITERATE(table->blob_entries->entrytable,
 			table->blob_entries->size,
 				add_entry, table);
@@ -850,6 +994,7 @@ err:
 
 static struct option connman_iptables_opts[] = {
 	{.name = "append",        .has_arg = 1, .val = 'A'},
+	{.name = "flush-chain",   .has_arg = 1, .val = 'F'},
 	{.name = "list",          .has_arg = 2, .val = 'L'},
 	{.name = "new-chain",     .has_arg = 1, .val = 'N'},
 	{.name = "delete-chain",  .has_arg = 1, .val = 'X'},
@@ -873,7 +1018,7 @@ int main(int argc, char *argv[])
 	struct xtables_match *xt_m;
 	struct xtables_target *xt_t;
 	char *table_name, *chain, *new_chain, *match_name, *target_name;
-	char *delete_chain;
+	char *delete_chain, *flush_chain;
 	int c;
 	size_t size;
 	gboolean dump, invert, delete;
@@ -884,18 +1029,22 @@ int main(int argc, char *argv[])
 	invert = FALSE;
 	delete = FALSE;
 	table_name = chain = new_chain = match_name = target_name = NULL;
-	delete_chain = NULL;
+	delete_chain = flush_chain = NULL;
 	table = NULL;
 	xt_m = NULL;
 	xt_t = NULL;
 
 	while ((c = getopt_long(argc, argv,
-	   "-A:L::N:X:j:i:m:o:t:", connman_iptables_globals.opts, NULL)) != -1) {
+	   "-A:F:L::N:X:j:i:m:o:t:", connman_iptables_globals.opts, NULL)) != -1) {
 		switch (c) {
 		case 'A':
 			chain = optarg;
 			break;
 
+		case 'F':
+			flush_chain = optarg;
+			break;
+
 		case 'L':
 			dump = true;
 			break;
@@ -927,7 +1076,11 @@ int main(int argc, char *argv[])
 			if (xt_t->init != NULL)
 				xt_t->init(xt_t->t);
 			connman_iptables_globals.opts =
-				xtables_merge_options(connman_iptables_globals.opts,
+				xtables_merge_options(
+#if XTABLES_VERSION_CODE > 5
+						     connman_iptables_globals.orig_opts,
+#endif
+						     connman_iptables_globals.opts,
 						     xt_t->extra_opts,
 						     &xt_t->option_offset);
 			if (connman_iptables_globals.opts == NULL)
@@ -953,7 +1106,11 @@ int main(int argc, char *argv[])
 				xt_m->init(xt_m->m);
 			if (xt_m != xt_m->next) {
 				connman_iptables_globals.opts =
-					xtables_merge_options(connman_iptables_globals.opts,
+					xtables_merge_options(
+#if XTABLES_VERSION_CODE > 5
+						connman_iptables_globals.orig_opts,
+#endif
+						connman_iptables_globals.opts,
 						xt_m->extra_opts,
 						&xt_m->option_offset);
 				if (connman_iptables_globals.opts == NULL)
@@ -1016,6 +1173,14 @@ int main(int argc, char *argv[])
 		goto commit;
 	}
 
+	if (flush_chain) {
+		printf("Flush chain %s\n", flush_chain);
+
+		connman_iptables_flush_chain(table, flush_chain);
+
+		goto commit;
+	}
+
 	if (dump) {
 		connman_iptables_dump(table);
 
